from __future__ import annotations

from pathlib import Path


def human_readable_size(
    size_bytes: int, units: str | None = None, return_float: bool = False
) -> str | float:
    """
    Convert bytes to a human-readable format (B, KB, MB, GB, TB).

    Parameters
    ----------
    size_bytes
        The size in bytes to convert.
    units
        The unit to return the size in (B, KB, MB, GB, TB).
    return_float
        If True, return the size as a float without units.

    Returns
    -------
    The size converted to the requested or most appropriate unit.
    """

    unit_list = ["B", "KB", "MB", "GB"]
    size_bytes = float(size_bytes)  # ensure float precision

    # if a specific unit is requested, convert directly to that unit
    if units and units in unit_list:
        unit_index = unit_list.index(units)
        size_bytes /= 1024**unit_index
        return size_bytes if return_float else f"{size_bytes:.2f} {units}"

    # otherwise, find the appropriate unit dynamically
    for unit in unit_list:
        if size_bytes < 1024:
            return size_bytes if return_float else f"{size_bytes:.2f} {unit}"
        size_bytes /= 1024

    # if it's beyond TB, return TB representation
    return size_bytes if return_float else f"{size_bytes:.2f} TB"


def _get_file_size(file_path: Path) -> int:
    """Get the size of a file in bytes."""

    return file_path.stat().st_size


def get_file_size(file_path: Path | str, **kwargs) -> str:
    """Get the size of a file in a human-readable format."""

    size_bytes = _get_file_size(Path(file_path))
    return human_readable_size(size_bytes, **kwargs)



from __future__ import annotations

import warnings
from pathlib import Path

import ase
import h5py
import numpy as np


class ASEH5Trajectory:
    def __init__(
        self,
        immutable: list[str] | None = None,
        mutable: list[str] | None = None,
        info_keys: list[str] | None = None,
    ) -> None:
        """
        Parameters
        ----------
        immutable
            List of immutable properties to write to the HDF5 file.
        mutable
            List of mutable properties to write to the HDF5 file.
        info_keys
            List of per-frame info keys to write to the HDF5 file.
        """

        default_immutable = ["numbers"]
        default_mutable = ["positions"]

        # merge defaults with provided lists
        immutable = (immutable or []) + default_immutable
        mutable = (mutable or []) + default_mutable

        # Validate keys
        self.immutable, self.mutable = validate_keys(immutable, mutable)
        self.info_keys = info_keys or []

    def write(
        self,
        atoms_list: list[ase.Atoms],
        filename: Path | str,
        compression: str = "gzip",
        float_type: str = "float32",
    ) -> None:
        """
        Write a list of ASE Atoms objects to an HDF5 file.

        Some properties are considered immutable, and are stored in the
        "immutable" group. These are properties that should not change between
        frames, such as atomic numbers. Mutable properties are stored in the
        "mutable" group, and are properties that can change between frames, such
        as atomic positions. The cell is considered mutable if it changes
        between frames, and immutable otherwise.

        Basic checks are performed to ensure that immutable properties do not
        change between frames and that mutable properties are present in all
        frames.

        Parameters
        ----------
        atoms_list
            List of ASE Atoms objects to write.
        filename
            Path to the HDF5 file to write to.
        compression
            Compression algorithm to use when writing the HDF5 file.

        Raises
        ------
        ValueError
            If a mutable property is missing in any frame.
        """

        filename = Path(filename)

        float_dtype = np.dtype(float_type)

        with h5py.File(filename, "w") as h5file:
            first_atoms = atoms_list[0]

            # handle immutable properties.
            for key in self.immutable:
                data = first_atoms.arrays.get(key, first_atoms.info.get(key))
                if data is not None:
                    check_immutable_consistency(atoms_list, key, data)
                    h5file.create_dataset(
                        f"immutable/{key}",
                        data=convert_dtype(data, float_dtype),
                        compression=compression,
                    )
                else:
                    raise ValueError(
                        f"Immutable property '{key}' missing in frame 1."
                    )

            # handle mutable properties.
            for key in self.mutable:
                frame_data = []
                for atoms in atoms_list:
                    data = atoms.arrays.get(key, atoms.info.get(key))
                    if data is None:
                        raise ValueError(
                            f"Mutable property '{key}' missing in a frame."
                        )
                    frame_data.append(convert_dtype(data, float_dtype))
                stacked_data = np.stack(frame_data, axis=0)
                h5file.create_dataset(
                    f"mutable/{key}", data=stacked_data, compression=compression
                )

            # get the cell data.
            cells = np.stack(
                [atoms.cell.array for atoms in atoms_list],
                axis=0,
                dtype=float_dtype,
            )
            if np.all(np.isclose(cells, cells[0])):
                h5file.create_dataset(
                    "immutable/cell", data=cells[0], compression=compression
                )
            else:
                h5file.create_dataset(
                    "mutable/cell", data=cells, compression=compression
                )

            # other per-frame info.
            for key in self.info_keys:
                data = np.array(
                    [atoms.info.get(key, np.nan) for atoms in atoms_list]
                )
                if np.any(np.isnan(data)):
                    warnings.warn(
                        f"Some frames missing '{key}' info.", stacklevel=2
                    )
                h5file.create_dataset(
                    f"info/{key}", data=data, compression=compression
                )

    def read(self, filename: Path | str) -> list[ase.Atoms]:
        """
        Read ASE Atoms objects from an HDF5 file.

        Parameters
        ----------
        filename
            Path to the HDF5 file to read from.

        Returns
        -------
        atoms_list
            List of ASE Atoms objects read from the file.
        """

        filename = Path(filename)

        atoms_list = []
        with h5py.File(filename, "r") as h5file:
            immutable_data = {
                key: decode_bytes(np.array(val))
                for key, val in h5file["immutable"].items()
            }
            mutable_data = {
                key: decode_bytes(np.array(val))
                for key, val in h5file["mutable"].items()
            }

            info_data = {}
            if "info" in h5file:
                info_data = {
                    key: decode_bytes(np.array(val))
                    for key, val in h5file["info"].items()
                }

            num_frames = next(iter(mutable_data.values())).shape[0]

            for i in range(num_frames):
                arrays = {
                    key: val[i]
                    for key, val in mutable_data.items()
                    if key != "cell"
                }
                arrays.update(immutable_data)

                cell = (
                    mutable_data.get("cell")[i]
                    if "cell" in mutable_data
                    else immutable_data.get("cell")
                )

                cell = cell if cell.sum() != 0 else None

                atoms = ase.Atoms(positions=arrays.pop("positions"), cell=cell)

                if cell is not None:
                    atoms.pbc = True

                for key, val in arrays.items():
                    atoms.arrays[key] = val

                for key, val in info_data.items():
                    atoms.info[key] = val[i]

                atoms_list.append(atoms)

        return atoms_list

    def __repr__(self) -> str:
        """A string representation of the ASEH5Trajectory object."""

        indent = " " * 4

        def format_keys(keys_set):
            return (
                ",\n".join(f"{indent * 2}{key}" for key in keys_set)
                or f"{indent * 2}<none>"
            )

        mutable_keys = format_keys(self.mutable)
        immutable_keys = format_keys(self.immutable)
        info_keys = format_keys(self.info_keys)

        ensemble = "NPT" if "cell" in self.mutable else "NVT"

        repr_lines = [
            f"{indent}immutable_keys=(\n{immutable_keys}\n{indent})",
            f"{indent}mutable_keys=(\n{mutable_keys}\n{indent})",
        ]

        if self.info_keys:
            repr_lines.append(
                f"info_keys=(\n{info_keys}\n{indent})",
            )

        return (
            f"ASEH5Trajectory(\n{indent}ensemble={ensemble},\n"
            + ",\n".join(repr_lines)
            + "\n)"
        )


########## HELPER FUNCTIONS ##########


def validate_keys(
    immutable: list[str] | None, mutable: list[str] | None
) -> tuple[set[str], set[str]]:
    """
    Ensure there are no repeated keys in both immutable and mutable. If
    'numbers' appears in mutable, remove it from immutable. If 'positions'
    appears in immutable, remove it from mutable.

    Parameters
    ----------
    immutable
        List of immutable properties.
    mutable
        List of mutable properties.

    Returns
    -------
    Validated sets of immutable and mutable properties.
    """

    immutable_set = set(immutable or [])
    mutable_set = set(mutable or [])

    # Ensure 'numbers' and 'positions' are correctly placed.
    if "numbers" in mutable_set:
        immutable_set.discard("numbers")

    if "positions" in immutable_set:
        mutable_set.discard("positions")

    # Now remove common keys
    common_keys = immutable_set & mutable_set
    if common_keys:
        raise ValueError(
            "Conflicting keys found in both immutable and mutable: "
            + ", ".join([f"'{x}'" for x in common_keys])
        )

    return immutable_set, mutable_set


def check_immutable_consistency(atoms_list, key, data) -> None:
    """
    Check if an immutable property changes between frames.
    """

    for atoms in atoms_list[1:]:
        new_data = atoms.arrays.get(key, atoms.info.get(key))
        if new_data is not None and not np.array_equal(data, new_data):
            warnings.warn(
                f"Immutable property '{key}' changes between frames.",
                stacklevel=2,
            )
            break


def convert_dtype(
    data: np.ndarray | float, float_dtype: np.dtype
) -> np.ndarray | float:
    """
    Convert data to the specified float dtype if it's of float type.
    For unicode strings, convert to fixed-length byte strings.
    Otherwise, return as-is.

    Parameters
    ----------
    data
        Input data (scalar or np.ndarray).
    float_dtype
        Target float dtype (e.g., np.float32 or np.dtype("float32")).

    Returns
    -------
    Converted data or original data.
    """

    if isinstance(data, np.ndarray):
        if np.issubdtype(data.dtype, np.floating):
            return data.astype(float_dtype)
        elif np.issubdtype(data.dtype, np.str_):
            itemsize = data.dtype.itemsize  # Number of bytes per char * length
            str_len = itemsize // np.dtype("U1").itemsize
            return data.astype(f"S{str_len}")
        else:
            return data
    elif isinstance(data, (float, np.floating)):
        return float_dtype(data)
    else:
        return data


def decode_bytes(data: np.ndarray) -> np.ndarray:
    """
    Decode a NumPy byte string array (dtype='S') to a Unicode string array
    (dtype='U').

    Parameters
    ----------
    data
        The array to decode.

    Returns
    -------
    Decoded array if input was byte string array; otherwise, return as-is.
    """

    if (
        isinstance(data, np.ndarray)
        and np.issubdtype(data.dtype, np.bytes_)
        and data.ndim > 0
        and data.dtype.kind == "S"
    ):
        return data.astype(str)

    return data



import pytest

from ase_hdf5.utils import _get_file_size, get_file_size, human_readable_size


@pytest.mark.parametrize(
    "size_bytes, expected_str, expected_float",
    [
        (500, "500.00 B", 500.00),
        (1023, "1023.00 B", 1023.00),
        (1024, "1.00 KB", 1.00),
        (1536, "1.50 KB", 1.50),
        (1048576, "1.00 MB", 1.00),
        (1073741824, "1.00 GB", 1.00),
        (1099511627776, "1.00 TB", 1.00),
    ],
)
def test_human_readable_size(size_bytes, expected_str, expected_float):
    """Test human_readable_size correctly converts bytes to readable format."""

    # Test default string output (auto unit selection)
    assert human_readable_size(size_bytes) == expected_str

    # Test float output
    assert human_readable_size(size_bytes, return_float=True) == pytest.approx(
        expected_float
    )


@pytest.mark.parametrize(
    "size_bytes, unit, expected_str, expected_float",
    [
        (500, "B", "500.00 B", 500.00),
        (500, "KB", "0.49 KB", 0.48828125),
        (1048576, "KB", "1024.00 KB", 1024.00),
        (1048576, "MB", "1.00 MB", 1.00),
        (1048576, "GB", "0.00 GB", 0.0009765625),
        (1073741824, "GB", "1.00 GB", 1.00),
        (1099511627776, "TB", "1.00 TB", 1.00),
    ],
)
def test_human_readable_size_specific_units(
    size_bytes, unit, expected_str, expected_float
):
    """Test human_readable_size conversion to a specific unit."""

    # Test string output for the requested unit
    assert human_readable_size(size_bytes, units=unit) == expected_str

    # Test float output for the requested unit
    assert human_readable_size(
        size_bytes, units=unit, return_float=True
    ) == pytest.approx(expected_float)


def test_get_file_size(tmp_path):
    """Test get_file_size by creating a temporary file and checking its size."""

    test_file = tmp_path / "test.txt"

    # Write 100 bytes to the file
    test_content = "a" * 100
    test_file.write_text(test_content)

    assert _get_file_size(test_file) == 100


def test_print_file_size(tmp_path):
    """Test print_file_size function captures correct output."""

    test_file = tmp_path / "test.txt"

    # Write 2048 bytes (2 KB).
    test_file.write_text("a" * 2048)
    file_size = get_file_size(test_file)

    assert "2.00 KB" in file_size



import ase
import numpy as np
import pytest

from ase_hdf5.core import (
    ASEH5Trajectory,
    convert_dtype,
    decode_bytes,
    validate_keys,
)


@pytest.fixture(params=["immutable_cell", "mutable_cell"])
def sample_atoms_list(request):
    """
    Return a list of 10 random ASE Atoms objects, with either immutable (NVT) or
    mutable (NPT) cell.
    """

    cell = np.random.rand(3, 3)
    numbers = np.random.randint(1, 10, size=10)
    framework = np.random.randint(1, 10, size=10)
    mol_id = np.random.randint(1, 10, size=10)

    atoms_list = []
    for _ in range(10):
        atoms = ase.Atoms(
            numbers=numbers,
            positions=np.random.rand(10, 3),
            cell=(
                np.random.rand(3, 3)
                if request.param == "mutable_cell"
                else cell
            ),
        )
        atoms.arrays["framework"] = framework
        atoms.arrays["mol-id"] = mol_id
        atoms_list.append(atoms)

    return atoms_list


def test_write_and_read(tmp_path, sample_atoms_list):
    """Tests writing/reading ASE Atoms objects to and from an HDF5 file."""

    test_file = tmp_path / "test.h5"

    traj = ASEH5Trajectory(
        immutable=["numbers", "framework", "mol-id"],
        mutable=["positions"],
    )

    traj.write(sample_atoms_list, test_file)
    read_atoms_list = traj.read(test_file)

    assert len(read_atoms_list) == len(sample_atoms_list)
    for orig, read in zip(sample_atoms_list, read_atoms_list):
        np.testing.assert_array_equal(orig.numbers, read.numbers)
        np.testing.assert_allclose(orig.positions, read.positions)
        np.testing.assert_allclose(orig.cell.array, read.cell.array)


def test_immutable_property_warning(tmp_path, sample_atoms_list):
    """Tests warning when an immutable property changes between frames."""

    test_file = tmp_path / "test.h5"
    sample_atoms_list[5].arrays["framework"] = np.random.randint(1, 10, size=10)

    traj = ASEH5Trajectory(immutable=["framework"], mutable=["positions"])

    with pytest.warns(
        UserWarning,
        match="Immutable property 'framework' changes between frames.",
    ):
        traj.write(sample_atoms_list, test_file)


def test_missing_mutable_property_error(tmp_path, sample_atoms_list):
    """Tests error when a mutable property is missing in a frame."""

    test_file = tmp_path / "test.h5"
    del sample_atoms_list[5].arrays["positions"]

    traj = ASEH5Trajectory(immutable=["numbers"], mutable=["positions"])

    with pytest.raises(
        ValueError, match="Mutable property 'positions' missing in a frame."
    ):
        traj.write(sample_atoms_list, test_file)


def test_empty_atoms_list(tmp_path):
    """Tests writing an empty list of ASE Atoms objects."""

    test_file = tmp_path / "test.h5"
    traj = ASEH5Trajectory()

    with pytest.raises(IndexError):
        traj.write([], test_file)


def test_info_keys(tmp_path, sample_atoms_list):
    """Tests writing/reading info keys to and from an HDF5 file."""

    test_file = tmp_path / "test.h5"
    for i, atoms in enumerate(sample_atoms_list):
        atoms.info["energy"] = float(i)

    traj = ASEH5Trajectory(info_keys=["energy"])
    traj.write(sample_atoms_list, test_file)
    read_atoms_list = traj.read(test_file)

    assert all(
        atoms.info["energy"] == float(i)
        for i, atoms in enumerate(read_atoms_list)
    )


def test_info_keys_warning(tmp_path, sample_atoms_list):
    """Tests warning when an info key is missing in a frame."""

    test_file = tmp_path / "test.h5"
    for i, atoms in enumerate(sample_atoms_list):
        atoms.info["energy"] = float(i) if i % 2 == 0 else np.nan

    traj = ASEH5Trajectory(info_keys=["energy"])

    with pytest.warns(UserWarning, match="Some frames missing 'energy' info."):
        traj.write(sample_atoms_list, test_file)


def test_immutable_property_handling(tmp_path):
    """Tests behavior when immutable properties are missing in some frames."""

    test_file = tmp_path / "test.h5"

    # Create an immutable cell
    cell = np.random.rand(3, 3)
    numbers = np.random.randint(1, 10, size=10)

    # Generate atoms list
    atoms_list = []
    for i in range(10):
        atoms = ase.Atoms(
            numbers=numbers,
            positions=np.random.rand(10, 3),
            cell=cell,
        )

        # Ensure some frames are missing an immutable property
        if i % 2 == 0:
            atoms.arrays["framework"] = np.random.randint(1, 10, size=10)
        else:
            atoms.arrays["framework"] = np.random.randint(20, 30, size=10)

        atoms_list.append(atoms)

    # Create a trajectory object where "framework" is an immutable property
    traj = ASEH5Trajectory(
        immutable=["numbers", "framework"], mutable=["positions"]
    )

    with pytest.warns(
        UserWarning,
        match="Immutable property 'framework' changes between frames.",
    ):
        traj.write(atoms_list, test_file)

    # Read back the trajectory
    read_atoms_list = traj.read(test_file)

    # Verify that the read data contains numbers and positions
    assert len(read_atoms_list) == len(atoms_list)
    for orig, read in zip(atoms_list, read_atoms_list):
        np.testing.assert_array_equal(orig.numbers, read.numbers)
        np.testing.assert_allclose(orig.positions, read.positions)

        # Check framework existence
        if "framework" in orig.arrays:
            assert "framework" in read.arrays
        else:
            assert "framework" not in read.arrays


def test_missing_immutable_property(tmp_path):
    """Tests behavior when an immutable property is missing in first frame."""

    test_file = tmp_path / "test.h5"

    # Create an immutable cell
    cell = np.random.rand(3, 3)
    numbers = np.random.randint(1, 10, size=10)
    framework = np.random.randint(1, 10, size=10)

    # Generate atoms list
    atoms_list = []
    for i in range(10):
        atoms = ase.Atoms(
            numbers=numbers,
            positions=np.random.rand(10, 3),
            cell=cell,
        )

        # Only set "framework" after the first frame (to trigger the ValueError)
        if i > 0:
            atoms.arrays["framework"] = framework

        atoms_list.append(atoms)

    # Create a trajectory object where "framework" is an immutable property
    traj = ASEH5Trajectory(
        immutable=["numbers", "framework"], mutable=["positions"]
    )

    # Expect a ValueError due to the missing immutable property in frame 0
    with pytest.raises(
        ValueError, match="Immutable property 'framework' missing in frame 1."
    ):
        traj.write(atoms_list, test_file)


def test_conflicting_keys():
    """
    Test if validate_keys raises an error for duplicate keys in immutable and
    mutable.
    """

    with pytest.raises(
        ValueError,
        match=(
            "Conflicting keys found in both immutable and mutable: 'shared_key'"
        ),
    ):
        validate_keys(["shared_key"], ["shared_key"])


def test_numbers_in_mutable():
    """Test if 'numbers' in mutable is removed from immutable."""

    immutable, mutable = validate_keys(["numbers"], ["numbers", "extra"])
    assert "numbers" not in immutable
    assert "extra" in mutable  # Ensure other mutable keys are still there


def test_positions_in_immutable():
    """Test if 'positions' in immutable is removed from mutable."""

    immutable, mutable = validate_keys(["positions", "extra"], ["positions"])
    assert "positions" not in mutable
    assert "extra" in immutable  # Ensure other immutable keys are still there


def test_read_no_cell_sets_no_pbc(tmp_path):
    """Ensure that if no cell is found, pbc is not set."""

    # create a valid file with all data *except* the cell.
    test_file = tmp_path / "no_cell.h5"
    atoms = ase.Atoms("H2O", positions=[[0, 0, 0], [0.8, 0, 0], [-0.4, 0.7, 0]])

    # write to file.
    traj = ASEH5Trajectory(immutable=["numbers"], mutable=["positions"])
    traj.write([atoms], test_file)

    # read back and test.
    atoms_list = traj.read(test_file)
    assert len(atoms_list) == 1
    assert not atoms_list[0].pbc.any()  # pbc should not be set.


########## TEST CONVERSIONS ##########


def test_convert_float_scalar():
    val = 3.14
    result = convert_dtype(val, np.float32)
    assert isinstance(result, np.float32)
    assert result == np.float32(3.14)


def test_convert_float_array():
    arr = np.array([1.0, 2.0, 3.0], dtype=np.float64)
    result = convert_dtype(arr, np.float32)
    assert result.dtype == np.float32
    np.testing.assert_allclose(result, arr, rtol=1e-6)


def test_convert_int_scalar():
    val = 42
    result = convert_dtype(val, np.float32)
    assert isinstance(result, int)  # Should remain unchanged
    assert result == 42


def test_convert_int_array():
    arr = np.array([1, 2, 3], dtype=np.int32)
    result = convert_dtype(arr, np.float32)
    assert result.dtype == np.int32  # Should remain unchanged
    np.testing.assert_array_equal(result, arr)


def test_convert_string_array():
    arr = np.array(["H", "C", "O"])
    result = convert_dtype(arr, np.float32)
    assert result.dtype.type is np.bytes_
    expected = np.array([b"H", b"C", b"O"])
    np.testing.assert_array_equal(result, expected)


def test_convert_object_array():
    arr = np.array([1, "H", 3.0], dtype=object)
    result = convert_dtype(arr, np.float32)
    assert result.dtype == object
    assert result.tolist() == arr.tolist()


def test_convert_empty_float_array():
    arr = np.array([], dtype=np.float64)
    result = convert_dtype(arr, np.float32)
    assert result.dtype == np.float32
    assert result.size == 0


def test_convert_empty_int_array():
    arr = np.array([], dtype=np.int64)
    result = convert_dtype(arr, np.float32)
    assert result.dtype == np.int64
    assert result.size == 0


def test_convert_already_correct_dtype():
    arr = np.array([1.0, 2.0, 3.0], dtype=np.float32)
    result = convert_dtype(arr, np.float32)
    assert result.dtype == np.float32
    np.testing.assert_array_equal(result, arr)


def test_decode_byte_string_array():
    byte_array = np.array([b"H", b"O", b"C"], dtype="S1")
    result = decode_bytes(byte_array)
    
    assert result.dtype.kind == "U"  # Unicode
    assert result.dtype.itemsize >= 1
    assert result.tolist() == ["H", "O", "C"]


def test_unicode_string_array_unchanged():
    str_array = np.array(["H", "O", "C"], dtype="U1")
    result = decode_bytes(str_array)
    
    assert result is str_array  # should return original
    assert result.dtype.kind == "U"
    assert result.tolist() == ["H", "O", "C"]


def test_numeric_array_unchanged():
    float_array = np.array([1.0, 2.0, 3.0], dtype=np.float32)
    result = decode_bytes(float_array)

    assert result is float_array
    np.testing.assert_array_equal(result, float_array)


def test_empty_byte_array():
    empty = np.array([], dtype="S1")
    result = decode_bytes(empty)

    assert result.dtype.kind == "U"
    assert result.size == 0


def test_non_ndarray_input():
    data = [b"H", b"O", b"C"]
    result = decode_bytes(data)

    assert result == data  # unchanged